package org.spagic.monitoring.jbpm;

import java.util.Date;
import java.util.List;
import java.util.Map;

import org.hibernate.Session;
import org.hibernate.Transaction;
import org.jbpm.context.exe.ContextInstance;
import org.jbpm.graph.def.Node;
import org.jbpm.graph.exe.Token;
import org.jbpm.graph.log.ProcessInstanceCreateLog;
import org.jbpm.graph.log.ProcessInstanceEndLog;
import org.jbpm.graph.log.TransitionLog;
import org.jbpm.logging.log.ProcessLog;
import org.spagic.metadb.base.ProcessInstanceStateConstants;
import org.spagic.metadb.base.TransitionStateConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.LogJournal;
import org.spagic.metadb.model.Process;
import org.spagic.metadb.model.ProcessInstance;
import org.spagic.metadb.model.Step;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.monitoring.core.MonitorServiceBase;
import org.spagic.monitoring.core.TransitionStateCache;
import org.spagic.monitoring.jbpm.message.JBPMMessage;
import org.spagic.monitoring.message.MessageIFace;
import org.spagic.workflow.api.jbpm.ProcessEngine;

public class MonitorServiceJBPM extends MonitorServiceBase {
	
	
	public MonitorServiceJBPM() {
	}
	
	public void traceMessage(MessageIFace message) {
		ProcessLog processLog = ((JBPMMessage)message).getProcessLog();
		traceLog(processLog);
	}
	
	protected void traceLog(ProcessLog processLog) {
		if (processLog instanceof TransitionLog) {
			processTransitionLog((TransitionLog)processLog);
			return;
		}
				
		if (processLog instanceof ProcessInstanceCreateLog) {
			// Creation of a process instance
			processProcessInstanceCreateLog((ProcessInstanceCreateLog)processLog);
			return;
		}
		
		if (processLog instanceof ProcessInstanceEndLog) {
			// Termination of a process instance
			processProcessInstanceEndLog((ProcessInstanceEndLog)processLog);
			return;
		}
	}

	/**
	 * Manage the ProcessInstanceCreateLog creating the process instance, only for subprocesses.
	 * @param processLog
	 */
	private void processProcessInstanceCreateLog(ProcessInstanceCreateLog processLog) {
		
		// A quale istanza di processo di riferisce questo log ?
		Token token = processLog.getToken();
		org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance = token.getProcessInstance();
		Token superProcessToken = jBPMProcessInstance.getSuperProcessToken();
				
		if (superProcessToken == null) {
			// This is a standalone process, already monitored
			return;
		}
		
		// Proceed only for subprocesses
		org.jbpm.graph.exe.ProcessInstance jBPMParentProcessInstance = superProcessToken.getProcessInstance();
		
					
		Session aSession = null;
		Transaction tx = null;
		try {
			// Retrieve current session
			aSession = getSession();
			tx = aSession.beginTransaction();
			
			// Create the process linked to the parent
			org.spagic.metadb.model.ProcessInstance spagicProcessInstance = ProcessEngine.createSpagicProcess(jBPMProcessInstance, jBPMParentProcessInstance, aSession);

			
			
			// In the main process, we have to associate to the current transition the id of the subprocess
			// 1. Retrieve Spagic parent process
			org.spagic.metadb.model.ProcessInstance spagicParentProcessInstance = ProcessEngine.getSpagicProcessInstanceFromJBPMProcessInstance(jBPMParentProcessInstance, aSession);
			// 2. Retrieve the parent current node name
			String parentNodeName = superProcessToken.getNode().getName();
			Step parentStep = getSpagicStepFromJBPMNode(spagicParentProcessInstance.getProcess(), parentNodeName, aSession);
			// 3. Retrieve the last transition written by the superprocess: it should be the
			// right one to link to the subprocess
			List<TransitionInstance> transitionInstances = 
				AuditDBUtils.getTransitionInstancesByProcessInstanceIdAndTargetStep(aSession, 
						spagicParentProcessInstance.getIdProcessInstance(), parentStep.getIdStep());
			if (transitionInstances.size() != 0) {
				TransitionInstance parentTransitionInstance = transitionInstances.get(transitionInstances.size() - 1);
				parentTransitionInstance.setIdExtProcessInstance(spagicProcessInstance.getIdProcessInstance().toString());
				aSession.save(parentTransitionInstance);
			}
			
			// Create process instance global variables  <<<---NOT POSSIBLE IN MONITORING
//			SpagicJbpmEngineUtils.createSpagicVariables(spagicProcessInstance, jBPMProcessInstance);
			
			aSession.flush();
			tx.commit();
		} catch (Exception e) {
			tx.rollback();
			e.printStackTrace();
        	log.error("traceExchange ", e);
		} finally {
			closeSession(aSession);
		}		
	}

	/**
	 * Manage the ProcessInstanceEndLog closing the current transition instance and process instance.
	 * @param processLog
	 */
	private void processProcessInstanceEndLog(ProcessInstanceEndLog processLog) {
		
		// A quale istanza di processo di riferisce questo log ?
		Token token = processLog.getToken();
		org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance = token.getProcessInstance();
					
		Session aSession = null;
		Transaction tx = null;
		try {
			// Retrieve current session
			aSession = getSession();
			tx = aSession.beginTransaction();

			// Retrieve Spagic process instance
			String correlationID = ProcessEngine.createSpagicCorrelationId(jBPMProcessInstance);
			ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceByIdCorrelation(aSession, correlationID);
			Process spagicProcess = spagicProcessInstance.getProcess();
			
			
			// The process is be finished: update its status
			spagicProcessInstance.setState(ProcessInstanceStateConstants.PROCESS_DONE);

			Date endDate = jBPMProcessInstance.getEnd();
			Date startDate = spagicProcessInstance.getStartdate();
			spagicProcessInstance.setEnddate(endDate);
			spagicProcessInstance.setExecutionTime(endDate.getTime() - startDate.getTime());
			
			aSession.save(spagicProcessInstance);
			
			
			
			// Retrieve the end node
			Node endNode = jBPMProcessInstance.getRootToken().getNode();
			String endNodeName = endNode.getName();
			Date processEndDate = jBPMProcessInstance.getRootToken().getEnd();
			
			// Retrieve Spagic target steps
			Step targetStep = getSpagicStepFromJBPMNode(spagicProcess, endNodeName, aSession);
			

			
			
			List<TransitionInstance> transitionInstances = AuditDBUtils.getTransitionInstancesByProcessInstanceIdAndTargetStep(aSession, spagicProcessInstance.getIdProcessInstance(), targetStep.getIdStep());
			for (TransitionInstance previousTransitionInstance : transitionInstances) {
				
				previousTransitionInstance.setTransitionEnd(processEndDate);
				previousTransitionInstance.setTransitionExecutionTime(previousTransitionInstance.getTransitionEnd().getTime() - previousTransitionInstance.getTransitionStart().getTime());
				previousTransitionInstance.setTransitionState(TransitionStateCache.getTransitionState(TransitionStateConstants.JBPM_DONE_ACTIVITY));
				
				aSession.save(previousTransitionInstance);
				
				// Trace all variables
				traceVariables(jBPMProcessInstance.getContextInstance(), aSession, previousTransitionInstance);
			}

			
			aSession.flush();
			tx.commit();
		} catch (Exception e) {
			tx.rollback();
        	log.error("traceExchange ", e);
//        	!!!
		} finally {
			closeSession(aSession);
		}		
	}
	
	
	private void processTransitionLog(TransitionLog transitionLog) {
				
		Token token = transitionLog.getToken();
		org.jbpm.graph.exe.ProcessInstance jBPMProcessInstance = token.getProcessInstance();
		Node sourceNode = transitionLog.getSourceNode();
		Node destinationNode = transitionLog.getDestinationNode();
					
		Session aSession = null;
		Transaction tx = null;
		try {
			// Retrieve current session
			aSession = getSession();
			tx = aSession.beginTransaction();
			
			
			// Suppose that the process definition is like the following one:
			//
			//      A ----> B ----> C -----> D
			//                  ^
			//
			// and that we are processing the transition B ---> C
			// 
			// To management of this log should:
			// 1. Create the transition instance B ---> C in state "available"
			// 2. Retrieves all transition instances for current process instance
			//    that have B as target step, and close them.
			//    In this case close transition instance A ----> B

			// Retrieve Spagic process instance
			String correlationID = ProcessEngine.createSpagicCorrelationId(jBPMProcessInstance);
			
			
			ProcessInstance spagicProcessInstance = AuditDBUtils.getProcessInstanceByIdCorrelation(aSession, correlationID);
			Process spagicProcess = spagicProcessInstance.getProcess();
				
			// Retrieve Spagic steps
			Step sourceStep = getSpagicStepFromJBPMNode(spagicProcess, sourceNode.getName(), aSession);
			Step targetStep = getSpagicStepFromJBPMNode(spagicProcess, destinationNode.getName(), aSession);
			
			
			
			// 1. Create the transition instance B ---> C in state "available"
			TransitionInstance transitionInstance = new TransitionInstance();
			transitionInstance.setProcessInstance(spagicProcessInstance);
			transitionInstance.setIdMessage(getMessageID(transitionLog));
			transitionInstance.setTransitionStart(transitionLog.getDate());
			transitionInstance.setTransitionState(TransitionStateCache.getTransitionState(TransitionStateConstants.JBPM_AVAILABLE_ACTIVITY));
			transitionInstance.setStepByIdStepSource(sourceStep);
			transitionInstance.setStepByIdStepTarget(targetStep);
			aSession.save(transitionInstance);
			
			// Trace all variables
			traceVariables(jBPMProcessInstance.getContextInstance(), aSession, transitionInstance);
			
			
			
			
			// 2. Retrieves all transition instances for current process instance
			//    that have B as target step, and close them.
			//    In this case close transition instance A ----> B

			List<TransitionInstance> transitionInstances = AuditDBUtils.getTransitionInstancesByProcessInstanceIdAndTargetStep(aSession, spagicProcessInstance.getIdProcessInstance(), sourceStep.getIdStep());
			for (TransitionInstance previousTransitionInstance : transitionInstances) {
				
				previousTransitionInstance.setTransitionEnd(transitionLog.getDate());
				previousTransitionInstance.setTransitionExecutionTime(previousTransitionInstance.getTransitionEnd().getTime() - previousTransitionInstance.getTransitionStart().getTime());
				previousTransitionInstance.setTransitionState(TransitionStateCache.getTransitionState(TransitionStateConstants.JBPM_DONE_ACTIVITY));
				
				aSession.save(previousTransitionInstance);
				
				// Trace all variables
				traceVariables(jBPMProcessInstance.getContextInstance(), aSession, previousTransitionInstance);
			}
			
			
			aSession.flush();
			tx.commit();
		} catch (Exception e) {
			tx.rollback();
			e.printStackTrace();
        	log.error("traceExchange ", e);
        	
		} finally {
			closeSession(aSession);
		}
		
	}

	
	/**
	 * Trace all process variables.
	 * @param contextInstance JBPM context instance
	 * @param aSession Hibernate session
	 * @param transitionInstance Current transition instance
	 */
	private void traceVariables(ContextInstance contextInstance, Session aSession, TransitionInstance transitionInstance) {
		Map variables = contextInstance.getVariables();
		if (variables != null) {
			StringBuffer message = new StringBuffer();
			
			for (Object variableName : variables.keySet()) {
				Object variableValue = variables.get(variableName);
				message.append(variableName);
				message.append("=");
				message.append(variableValue);
				message.append("\r\n");
			}
			
			if (message.length() != 0) {
				LogJournal messageInfo = new LogJournal();
				messageInfo.setTransitionInstance(transitionInstance);
				messageInfo.setTransitionState(transitionInstance.getTransitionState());
				messageInfo.setDatelog((transitionInstance.getTransitionEnd() != null) ? transitionInstance.getTransitionEnd() : transitionInstance.getTransitionStart());
				messageInfo.setLog(message.toString());

				aSession.save(messageInfo);
			}
		}
	}
	
	public Step getSpagicStepFromJBPMNode(Process spagicProcess, String nodeName, Session aSession) {
		String processName = spagicProcess.getName();
		String processNamespace = spagicProcess.getNamespace();
		String stepName = ProcessEngine.getStepLocalNamePart(processName, spagicProcess.getProcessVersion(), nodeName);
		Step targetStep = AuditDBUtils.getStepByName(aSession, processNamespace, stepName);
		return targetStep;
	}
		
	public String getMessageID(TransitionLog transitionLog) {
		String messageID = ProcessEngine.JBPM_PREFIX
				+ ProcessEngine.MESSAGE_ID_PREFIX_TOKEN + transitionLog.getToken().getId()
				+ ProcessEngine.MESSAGE_ID_PREFIX_START_NODE + transitionLog.getSourceNode().getId()
				+ ProcessEngine.MESSAGE_ID_PREFIX_DESTINATION_NODE + transitionLog.getDestinationNode().getId()
				+ ProcessEngine.MESSAGE_ID_PREFIX_TIMESTAMP + transitionLog.getDate().getTime();
		return messageID;
	}
	
	

	protected Session getSession() {
		Session aSession = HibernateUtil.getSessionFactory().openSession();
//		Session aSession = HibernateUtil.getSessionFactory().getCurrentSession();
		return aSession;
	}
		
	protected void closeSession(Session aSession) {
		if (aSession != null && aSession.isOpen()) {
			aSession.close();
		}
	}
	
}
